(********************************************)
(*            hresolve ver 1.0              *)
(*              Chung-Kil Hur               *)
(*       Preuves Programmes Systemes        *)
(********************************************)

(** 
    * hresolve_core and hget_evar

    Author: Chung-Kil Hur

    Institution: Laboratoire Preuves Programmes Systemes, CNRS & Universite Paris Diderot *)

open Pp
open Util
open Names
open Term
open Termops
open Proof_type
open Tacmach
open Tactics
open Rawterm
open Refiner
open Environ

(**
   hresolve_core
*)

let map_rawdecl_left_to_right f (na,k,obd,ty) = 
  let comp1 = Option.map f obd in
  let comp2 = f ty in
  (na,k,comp1,comp2)

let map_rawconstr_left_to_right f = function
  | RApp (loc,g,args) -> 
      let comp1 = f g in 
      let comp2 = Util.list_map_left f args in 
      RApp (loc,comp1,comp2)
  | RLambda (loc,na,bk,ty,c) -> 
      let comp1 = f ty in 
      let comp2 = f c in 
      RLambda (loc,na,bk,comp1,comp2)
  | RProd (loc,na,bk,ty,c) -> 
      let comp1 = f ty in 
      let comp2 = f c in 
      RProd (loc,na,bk,comp1,comp2)
  | RLetIn (loc,na,b,c) -> 
      let comp1 = f b in 
      let comp2 = f c in 
      RLetIn (loc,na,comp1,comp2)
  | RCases (loc,sty,rtntypopt,tml,pl) ->
      let comp1 = Option.map f rtntypopt in 
      let comp2 = Util.list_map_left (fun (tm,x) -> (f tm,x)) tml in
      let comp3 = Util.list_map_left (fun (loc,idl,p,c) -> (loc,idl,p,f c)) pl in
      RCases (loc,sty,comp1,comp2,comp3)
  | RLetTuple (loc,nal,(na,po),b,c) ->
      let comp1 = Option.map f po in
      let comp2 = f b in
      let comp3 = f c in 
      RLetTuple (loc,nal,(na,comp1),comp2,comp3)
  | RIf (loc,c,(na,po),b1,b2) ->
      let comp1 = Option.map f po in
      let comp2 = f b1 in
      let comp3 = f b2 in
      RIf (loc,f c,(na,comp1),comp2,comp3)
  | RRec (loc,fk,idl,bl,tyl,bv) ->
      let comp1 = Array.map (Util.list_map_left (map_rawdecl_left_to_right f)) bl in
      let comp2 = Array.map f tyl in
      let comp3 = Array.map f bv in
      RRec (loc,fk,idl,comp1,comp2,comp3)
  | RCast (loc,c,k) -> 
      let comp1 = f c in
      let comp2 = match k with CastConv (k,t) -> CastConv (k, f t) | x -> x in
      RCast (loc,comp1,comp2)
  | (RVar _ | RSort _ | RHole _ | RRef _ | REvar _ | RPatVar _ | RDynamic _) as x -> x

let subst_var_with_hole occ tid t = 
  let occref = if occ > 0 then ref occ else error_invalid_occurrence [occ] in
  let locref = ref 0 in
  let rec substrec = function
    | RVar (_,id) as x -> 
        if id = tid 
        then (decr occref; if !occref = 0 then x
                           else (incr locref; RHole (Ploc.make !locref 0 (0,0),Evd.QuestionMark(Evd.Define true))))
        else x
    | c -> map_rawconstr_left_to_right substrec c in
  let t' = substrec t
  in
  if !occref > 0 then error_invalid_occurrence [occ] else t'

let subst_hole_with_term occ tc t =
  let locref = ref 0 in
  let occref = ref occ in
  let rec substrec = function
    | RHole (_,Evd.QuestionMark(Evd.Define true)) -> 
        decr occref; if !occref = 0 then tc
                     else (incr locref; RHole (Ploc.make !locref 0 (0,0),Evd.QuestionMark(Evd.Define true)))
    | c -> map_rawconstr_left_to_right substrec c
  in
  substrec t

let clearbody id env =
  let rec aux _ = function
  | (id',Some c,t) when id = id' -> push_named (id,None,t)
  | d -> push_named d in
  fold_named_context aux env ~init:(reset_context env)

let hResolve id c occ t gl = 
  let occ = 
    match occ with
    | ArgArg n -> n
    | ArgVar _ -> error "hresolve_core called with a non closed argument." in
  let sigma = project gl in 
  let env = clearbody id (pf_env gl) in
  let env_ids = ids_of_context env in
  let env_names = names_of_rel_context env in
  let c_raw = Detyping.detype true env_ids env_names c in 
  let t_raw = Detyping.detype true env_ids env_names t in 
  let rec resolve_hole t_hole =
    (* Pp.msg (Printer.pr_lrawconstr_env env t_hole); print_string "\n"; *)
    try 
      Pretyping.Default.understand sigma env t_hole
    with 
    | Ploc.Exc (loc,Pretype_errors.PretypeError (_, Pretype_errors.UnsolvableImplicit _)) -> 
        (* print_int (Ploc.line_nb loc); print_string "\n"; *)
        resolve_hole (subst_hole_with_term (Ploc.line_nb loc) c_raw t_hole)
  in
  let t_constr = resolve_hole (subst_var_with_hole occ id t_raw) in
  let t_constr_type = Retyping.get_type_of env sigma t_constr in
  (* Pp.msg (Printer.pr_lconstr t_constr); print_string "\n";  *)
  change_in_concl None (mkLetIn (Anonymous,t_constr,t_constr_type,pf_concl gl)) gl

let hResolve_auto id c t gl =
  let rec resolve_auto n = 
    try
      hResolve id c (ArgArg n) t gl
    with
    | UserError (arg1,arg2) -> raise (UserError (arg1,arg2))
    | _ -> resolve_auto (n+1)
  in
  resolve_auto 1

TACTIC EXTEND hresolve_core
| [ "hresolve_core" "(" ident(id) ":=" constr(c) ")" "at" int_or_var(occ) "in" constr(t) ] -> [ hResolve id c occ t ]
| [ "hresolve_core" "(" ident(id) ":=" constr(c) ")" "in" constr(t) ] -> [ hResolve_auto id c t ]
END

(**
   hget_evar
*)

open Evar_refiner
open Tacexpr
open Evd
open Sign

let evar_list evc c =
  let rec evrec acc c =
    match kind_of_term c with
    | Evar (n, _) when Evd.mem evc n -> c :: acc
    | _ -> fold_constr evrec acc c
  in
    evrec [] c

let hget_evar n gl =
  let n = 
    match n with
    | ArgArg n -> n
    | ArgVar _ -> error "hget_evar called with a non closed argument." in
  let sigma = gl.sigma in
  let env = pf_env gl in
  let evl = evar_list sigma gl.it.evar_concl in
    if List.length evl < n then
      error "not enough uninstantiated existential variables";
    if n <= 0 then error "Incorrect existential variable index.";
    let ev = List.nth evl (n-1) in
    let ev_type = Retyping.get_type_of env sigma ev in
    change_in_concl None (mkLetIn (Anonymous,ev,ev_type,pf_concl gl)) gl

TACTIC EXTEND hget_evar
| [ "hget_evar" int_or_var(n) ] -> [ hget_evar n ]
END
